// lattice class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "Lattice.h"

#include <iostream>
#include <ecsb/ecsb.h>
#include <stdlib.h>

using namespace Simulation;
using namespace ECSB;

// constructor
Lattice::Lattice(unsigned long x, unsigned long y, unsigned long z, \
		double s)
{
	register unsigned long i, j, k;

	// set lattice dimensions
	xdim = x;
	ydim = y;
	zdim = z;

	// set lattice scale
	scale = s;

	// set lattice scale + site volume
	site_volume = scale * scale * scale;

	// set initial temperature
	temperature = 1.0;

	// allocate memory for + initialise lattice
	if(!(lattice = (LatticeSite ****) calloc(xdim, sizeof(LatticeSite ***)))) {
		cerr << "error: couldn't allocate lattice\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < xdim; i++) {
		if(!(lattice[i] = (LatticeSite ***) calloc(ydim, \
				sizeof(LatticeSite **)))) {
			cerr << "error: couldn't allocate lattice\n";
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < ydim; j++) {
			if(!(lattice[i][j] = (LatticeSite **) calloc(zdim, \
					sizeof(LatticeSite *)))) {
				cerr << "error: couldn't allocate lattice\n";
				exit(EXIT_FAILURE);
			}
			for(k = 0; k < zdim; k++) {
				if(!(lattice[i][j][k] = (LatticeSite *) malloc( \
						sizeof(LatticeSite)))) {
					cerr << "error: couldn't allocate lattice\n";
					exit(EXIT_FAILURE);
				}
				lattice[i][j][k]->x = i;
				lattice[i][j][k]->y = j;
				lattice[i][j][k]->z = k;
				lattice[i][j][k]->spin = 0;
				lattice[i][j][k]->type = 0;
			}
		}
	}

	// number of neighbours TODO: generalise for any lattice
	num_neighbours = 4;

	// allocate memory for vectors to calculate neighbours
	if(!(nxvector = (int *) calloc(neighbourhood_size, \
			sizeof(int)))) {
		cerr << "error: couldn't allocate neighbourhood xvectors\n";
		exit(EXIT_FAILURE);
	}
	if(!(nyvector = (int *) calloc(neighbourhood_size, \
			sizeof(int)))) {
		cerr << "error: couldn't allocate neighbourhood yvectors\n";
		exit(EXIT_FAILURE);
	}
	if(!(nzvector = (int *) calloc(neighbourhood_size, \
			sizeof(int)))) {
		cerr << "error: couldn't allocate neighbourhood zvectors\n";
		exit(EXIT_FAILURE);
	}

	// set neighbourhood vectors TODO: generalise for any lattice
	nxvector[0] = 1;
	nyvector[0] = 0;
	nzvector[0] = 0;
	nxvector[1] = 0;
	nyvector[1] = -1;
	nzvector[1] = 0;
	nxvector[2] = -1;
	nyvector[2] = 0;
	nzvector[2] = 0;
	nxvector[3] = 0;
	nyvector[3] = 1;
	nzvector[3] = 0;

	// allocate memory for temporary storage for site neighbours
	num_neighbours = 0;
	if(!(neighbours = (LatticeSite **) calloc(neighbourhood_size, \
			sizeof(LatticeSite *)))) {
		cerr << "error: couldn't allocate lattice neighbours\n";
		exit(EXIT_FAILURE);
	}
}

// destructor
Lattice::~Lattice()
{
	register unsigned long i, j, k;

	// free memory
	free(neighbours);
	free(nzvector);
	free(nyvector);
	free(nxvector);
	for(i = 0; i < xdim; i++) {
		for(j = 0; j < ydim; j++) {
			for(k = 0; k < zdim; k++)
				free(lattice[i][j][k]);
			free(lattice[i][j]);
		}
		free(lattice[i]);
	}
	free(lattice);
}

// get size of lattice neighbourhood
unsigned int Lattice::getNeighbourhoodSize(void) { return neighbourhood_size; }

// get volume of a single lattice site
double Lattice::getSiteVolume(void) { return site_volume; }

// set lattice temperature
void Lattice::setTemperature(double t) { temperature = t; }

// get lattice temperature
double Lattice::getTemperature(void) { return temperature; }

// get random site on lattice
LatticeSite *Lattice::getRandomSite(void)
{
	if(zdim == 1)
		// 2d lattice
		return lattice[randomNonnegativeInteger(xdim)] \
				[randomNonnegativeInteger(ydim)][0];
	else
		// 3d lattice
		return lattice[randomNonnegativeInteger(xdim)] \
				[randomNonnegativeInteger(ydim)] \
				[randomNonnegativeInteger(zdim)];
}

// calculate list of alien neighbours i.e. with different spin
unsigned int Lattice::calculateAlienNeighbours(LatticeSite *site)
{
	register unsigned int i;
	register unsigned long site_spin;
	register LatticeSite *neighbour;

	site_spin = site->spin;
	num_neighbours = 0;
	if(zdim == 1)
		// 2d lattice
		for(i = 0; i < neighbourhood_size; i++) {
			neighbour = \
					lattice[site->x + nxvector[i]][site->y + nyvector[i]][0];
			if(neighbour->spin != site_spin)
				neighbours[num_neighbours++] = neighbour;
		}
	else
		// 3d lattice
		for(i = 0; i < neighbourhood_size; i++) {
			neighbour = lattice[site->x + nxvector[i]][site->y + nyvector[i]] \
					[site->z + nzvector[i]];
			if(neighbour->spin != site_spin)
				neighbours[num_neighbours++] = neighbour;
		}

	return num_neighbours;
}

// select random alien neighbour (only use after calculateAlienNeighbours)
LatticeSite *Lattice::selectAlienNeighbour(void)
{
	return neighbours[randomNonnegativeInteger(num_neighbours)];
}

// flip site from one spin to another
void Lattice::flipSite(LatticeSite *from, LatticeSite *to)
{
	from->spin = to->spin;
	from->type = to->type;
}

// calculate energy to flip to one spin from another
double Lattice::calculateFlipEnergy(LatticeSite *from, LatticeSite *to)
{
	register unsigned int i;
	register unsigned long from_spin, to_spin;
	register LatticeSite *neighbour;
	register double from_energy, to_energy;

	// calculate current energy and flipped energy of lattice site
	from_spin = from->spin;
	to_spin = to->spin;
	from_energy = 0.0;
	to_energy = 0.0;
	if(zdim == 1)
		// 2d lattice
		for(i = 0; i < neighbourhood_size; i++) {
			neighbour = \
					lattice[from->x + nxvector[i]][from->y + nyvector[i]][0];
			if(neighbour->spin != from_spin)
				from_energy += 1.0;  // current energy
			if(neighbour->spin != to_spin)
				to_energy += 1.0;  // flipped energy
		}
	else
		// 3d lattice
		for(i = 0; i < neighbourhood_size; i++) {
			neighbour = lattice[from->x + nxvector[i]][from->y + nyvector[i]] \
					[from->z + nzvector[i]];
			if(neighbour->spin != from_spin)
				from_energy += 1.0;  // current energy
			if(neighbour->spin != to_spin)
				to_energy += 1.0;  // flipped energy
		}

	return (to_energy - from_energy);
}

// initialise lattice from p system
double Lattice::initialise(Psystem *ps)
{
	register unsigned long i, j, k;

	// clear lattice
	for(i = 0; i < xdim; i++)
		for(j = 0; j < ydim; j++)
			for(k = 0; k < zdim; k++)
				lattice[i][j][k]->spin = lattice[i][j][k]->type = 0;

	// assign spins and types to lattice
}